#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from dataclasses import dataclass

from entity_read import sql
from entity_read.entity import Entity
from entity_read.sql.atoms import Selectable

from .expression import Expression


@dataclass(frozen=True, kw_only=True, repr=False)
class Column(Expression):
    key: str

    def eval(self, entity_type: type[Entity], variables: dict[str, Selectable]) -> sql.atoms.column.Column:
        if (column_field := entity_type.get_key_to_column_field().get(self.key)) is None:
            raise KeyError(f"Not found column {self.key!r} in {entity_type.__name__}({entity_type.get_column_names()})")

        return sql.atoms.column.Column(key=self.key, type_=column_field.type)

    def shortcut(self) -> str:
        return f"col.{self.key}"



#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from dataclasses import dataclass

import entity_read.sql.inter.other_func
from entity_read import sql
from entity_read.entity import Entity
from .expression import Expression

key_to_function_factory = {
    "len": entity_read.sql.inter.other_func.FunctionGetSubqueryLength,
    "exists": entity_read.sql.inter.other_func.FunctionExists,
    "not_exists": entity_read.sql.inter.other_func.FunctionNotExists,
    "to_str": entity_read.sql.inter.other_func.FunctionCastToStr,
    "sub": entity_read.sql.inter.other_func.FunctionSub,
    "div": entity_read.sql.inter.other_func.FunctionDiv,
    "json_extract_text": entity_read.sql.inter.json_func.FunctionJsonExtractText,
    "json_has_key": entity_read.sql.inter.json_func.FunctionJsonHasKey,
}


@dataclass(frozen=True, kw_only=True, repr=False)
class Function(Expression):
    key: str
    args: list[Expression]

    @property
    def keep_subquery(self) -> bool:
        return self._get_factory().keep_subquery

    def _get_factory(self) -> type[sql.atoms.Function]:
        if (function_factory := key_to_function_factory.get(self.key)) is None:
            raise KeyError(f"Not found function {self.key!r}")
        return function_factory

    def eval(self, entity_type: type[Entity], variables: dict[str, sql.atoms.Selectable]) -> sql.atoms.Function:
        function_factory = self._get_factory()
        # noinspection PyArgumentList
        return function_factory(args=[arg.eval(entity_type, variables) for arg in self.args])

    def shortcut(self) -> str:
        return f"func.{self.key}({','.join([arg.shortcut() for arg in self.args])})"



#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from dataclasses import dataclass, field

from entity_read import sql
from entity_read.entity import Entity
from entity_read.sql.atoms import Selectable

from ._eval_to_selector import eval_to_selector
# from .variable import Variable
from .query_token import QueryToken
from .reference import Reference
from .expression import Expression
from .order import Order
from .relation import Relation


@dataclass(frozen=True, kw_only=True, repr=False)
class SubQuery(Expression):
    over: Relation | Reference
    attrs: list[Expression] | Expression = field(default_factory=list)
    vars: dict[str, Expression] = field(default_factory=dict)
    filters: list[Expression] = field(default_factory=list)
    searches: list[Expression] = field(default_factory=list)
    orders: list[Order] = field(default_factory=list)
    limit: int | None = None
    offset: int | None = None

    def eval(self, entity_type: type[Entity], variables: dict[str, Selectable]) -> sql.lower.LowerSelector:
        is_scalar = not isinstance(self.over, Relation) or self.over.get_is_scalar(entity_type)
        result = eval_to_selector(
            entity_type=self.over.get_related_entity(entity_type),
            local_column_to_remote=self.over.get_remote_column_name_to_local(entity_type),  # relatively to child
            my_step=self.over.key, rename=self.shortcut(), is_scalar=is_scalar,
            attrs=self.attrs, vars_=self.vars, filters=self.filters, searches=self.searches,
            orders=self.orders, limit=self.limit, offset=self.offset
        )
        return result

    def shortcut(self) -> str:
        parts = []
        parts += prepare_tokens('attrs', self.attrs)
        parts += [f'vars={",".join([f"{k}={v.shortcut()}" for k, v in self.vars.items()])}'] if self.vars else []
        parts += prepare_tokens('filters', self.filters)
        parts += prepare_tokens('searches', self.searches)
        parts += prepare_tokens('orders', self.orders)
        parts += [f'limit={self.limit}'] if self.limit else []
        parts += [f'offset={self.offset}'] if self.offset else []
        return f'{self.over.shortcut()}.subquery({",".join(parts)})'


def prepare_tokens(key: str, vals: list[QueryToken] | None) -> list[str]:
    if isinstance(vals, list):
        return [f'{key}=[{",".join([val.shortcut() for val in vals])}]'] if vals else []
    elif isinstance(vals, QueryToken):
        return [f'{key}={vals.shortcut()}']
    else:
        return []
